/* a proof of concept for DTMF detection with the goertzel algo

   written 1996 by christian mock <cm@tahina.priv.at>

   This software may be used and distributed according to the terms
   of the GNU General Public License, incorporated herein by reference.


   if you're looking for an optimized integer-only implementation, get
   the linux kernel sources and look in drivers/isdn/isdn_audio.c --
   the kernel code is based on this code.


   The implementation of the goertzel algorithm is based on the paper
   
   Performance Evaluation and Real-Time Implementation of Subspace,
   Adaptive, and DFT Algorithms for Multi-tone Detection by
   Gner Arslan, Brian L. Evans, F. Ayhan Sakarya and Jos Luis Pino
   http://ptolemy.eecs.berkeley.edu/papers/96/dtmf_ict/

   $Id: goertzel.c,v 1.7 2004/07/23 17:17:03 cm Exp $

   */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <stdio.h>

#define NPOINTS 205		/* samples/run */
#define SAMPLERATE 8000
#define NCOEFF 16		/* nr of frequencies to be analyzed */

#define DTMF_TRESH 100000	/* min. volume for DTMF */
#define SILENCE_TRESH 100	/* max. volume for silence */
#define H2_TRESH    20000	/* 2nd harmonic */
#define DTMF_DUR       25	/* nr of samples to detect a DTMF tone */

#define AMP_BITS  9		/* bits in amplitude -- can't be 16,
				   since we'd get integer overflows */

#define LOGRP 0
#define HIGRP 1

int mulaw[256];
int cos2pik[NCOEFF];		/* 2 * cos(2 * PI * k / N), precalculated */

typedef struct {
  int freq;			/* frequency */
  int grp;			/* low/high group */
  int k;			/* k */
  int k2;			/* k for 1st harmonic */
} dtmf_t;



dtmf_t dtmf_tones[8] = {
  { 697, LOGRP, 0, 0 },
  { 770, LOGRP, 0, 0 },
  { 852, LOGRP, 0, 0 },
  { 941, LOGRP, 0, 0 },
  {1209, HIGRP, 0, 0 },
  {1336, HIGRP, 0, 0 },
  {1477, HIGRP, 0, 0 },
  {1633, HIGRP, 0, 0 }
};

char dtmf_matrix[4][4] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};

void init(void)
{
   int i, kp, k;


   /* 
    *  Set up mu-law conversion table
    */
   for(i=0;i<256;i++)
   {
      short mu,e,f,y;
      short etab[]={0,132,396,924,1980,4092,8316,16764};

      mu=255-i;
      e=(mu&0x70)/16;
      f=mu&0x0f;
      y=f*(1<<(e+3));
      y=etab[e]+y;
      if(mu&0x80)
	 y=-y;
      mulaw[i]=(y >> (15 - AMP_BITS));
   }

   /* precalculate 2 * cos (2 PI k / N) */
   kp = 0;
   for(i = 0; i < 8; i++) {
     k = (int)rint((double)dtmf_tones[i].freq * NPOINTS / SAMPLERATE);
     cos2pik[kp] = 2 * 32768 * cos(2 * PI * k / NPOINTS);
     dtmf_tones[i].k = kp++;
     
     k = (int)rint((double)dtmf_tones[i].freq * 2 * NPOINTS / SAMPLERATE);
     cos2pik[kp] = 2 * 32768 * cos(2 * PI * k / NPOINTS);
     dtmf_tones[i].k2 = kp++;
   }
}

int get_buffer(int fd, int buffer[], int samples) {
  int nread, i;
  int *ptr;
  unsigned char *ptr2;
  
  char readbuffer[NPOINTS];
  
  nread = read(fd, readbuffer, samples);
  if(nread < samples) {
    for(i = nread; i < samples; i++)
      readbuffer[i] = 0;
    nread = 0;
  }

  ptr = buffer;
  ptr2 = readbuffer;
  
  for(i = 0; i < samples; i++)
    *(ptr++) = mulaw[*(ptr2++)];

  return nread;
}


int goertzel(int N, int K, int *sample, int *cos2pik, int *result) {
  int sk, sk1, sk2;
  int k, n;

  for(k = 0; k < K; k++) {
    sk = sk1 = sk2 = 0.0;
    for(n = 0; n < N; n++) {
      sk = sample[n] + ((cos2pik[k] * sk1) >> 15) - sk2;
      sk2 = sk1;
      sk1 = sk;
    }
    result[k] =
      ((sk * sk) >> AMP_BITS) - ((((cos2pik[k] * sk) >> 15) * sk2) >> AMP_BITS) +
      ((sk2 * sk2) >> AMP_BITS);
  }
  return K;
}


void main(int argc, char **argv) {

  int buffer[NPOINTS];
  int result[NCOEFF];
  int fd;
  int i, j;
  int grp[2], silence;
  char what, last, dtmf_len;
  
  init();

  fd = open(argv[1], O_RDONLY);

  last = ' ';
  while(get_buffer(fd, buffer, NPOINTS)) {
    goertzel(NPOINTS, NCOEFF, buffer, cos2pik, result);

    grp[0] = grp[1] = -2;
    silence = 0;
    for(i = 0; i < 8; i++)
      if((result[dtmf_tones[i].k] > DTMF_TRESH) &&
	 (result[dtmf_tones[i].k2] < H2_TRESH))
	grp[dtmf_tones[i].grp] = (grp[dtmf_tones[i].grp] == -2  ? i : -1);
      else
	if((result[dtmf_tones[i].k] < SILENCE_TRESH) &&
	   (result[dtmf_tones[i].k2] < SILENCE_TRESH))
	  silence++;
    if(silence == 8)
      what = ' ';
    else {
      if((grp[LOGRP] >= 0) && (grp[HIGRP] >= 0)) {
	what = dtmf_matrix[grp[LOGRP]][grp[HIGRP] - 4];
	if(last != ' ' && last != '.')
	  last = what;		/* min. 1 period of non-DTMF between tones */
	/*	printf("\n");
	for(j = 0; j < 16; j++, j++)
	  printf("%10d %10d\n", result[j], result[j+1]); */
      } else
	what = '.';
    }
    if(what != last)
      printf("%c", what); 
    last = what;
  }
  printf("|\n");
  close(fd);
}

/*

  $Log: goertzel.c,v $
  Revision 1.7  2004/07/23 17:17:03  cm
  prepare for publication (license, references, ...)

  Revision 1.6  1997/03/07 08:54:14  cm
  translation to english

  Revision 1.5  1996/06/02 21:03:47  cm
  *** empty log message ***

 * Revision 1.4  1996/05/31  23:19:58  cm
 * fast fertig :-)
 *
 * Revision 1.3  1996/05/31  22:31:16  cm
 * int-version, mit einschraenkung  der amplitude auf 9 bits wegen
 * overflow-problemen.
 *
 * Revision 1.2  1996/05/31  21:52:04  cm
 * von double auf int bzw. long long -- long long ist
 * geschwindigkeitsmaessig alles andere als die erleuchtung.
 *
 * Revision 1.1  1996/05/31  21:25:22  cm
 * Initial revision
 *

*/
